package meerkat;

import com.google.protobuf.ByteString;
import com.google.protobuf.InvalidProtocolBufferException;
import meerkat.bulletinboard.SimpleBulletinBoardClient;
import meerkat.comm.CommunicationException;
import meerkat.crypto.Encryption;
import meerkat.protobuf.BulletinBoardAPI;
import meerkat.protobuf.Crypto;
import util.AccurateTimestamp;
import util.CollectionMessagesUtils;
import util.RegistryTags;

import java.io.IOException;
import java.text.ParseException;
import java.util.*;

/**
 * Created by Vladimir Eliezer Tokarev on 1/8/2016.
 * Gives the ability to synchronously manage voters information
 */
public class SimpleRegistry  extends RegistryInstance{

    protected Encryption signatory;

    protected SimpleBulletinBoardClient communicator;

    public SimpleRegistry(Encryption signatory, SimpleBulletinBoardClient communicator) {
        this.signatory = signatory;
        this.communicator = communicator;
    }

    /**
     * Creates BulletinBoardMessage with signed basicMessage and UnsignedBulletinBoardMessage that contains the basic message
     * @param basicMessage BasicMessage
     * @return BulletinBoardAPI.BulletinBoardMessage
     */
    private BulletinBoardAPI.BulletinBoardMessage CreateBulletinBoardMessage(BulletinBoardAPI.UnsignedBulletinBoardMessage basicMessage) {
        try {
            BulletinBoardAPI.BulletinBoardMessage.Builder bulletinBoardMessage =
                    BulletinBoardAPI.BulletinBoardMessage.newBuilder();

            Crypto.RerandomizableEncryptedMessage encryptedMessage =
                        signatory.encrypt(basicMessage, signatory.generateRandomness(new Random()));
            Crypto.Signature.Builder messageSignature = Crypto.Signature.newBuilder();
            messageSignature.mergeFrom(encryptedMessage);

            bulletinBoardMessage.setMsg(basicMessage);

            bulletinBoardMessage.addSig(messageSignature);

            return bulletinBoardMessage.build();
        } catch (IOException e) {
           return null;
        }
    }

    /**
     * Gets messages that have the wanted id tag and have or the ADD_TO_GROUP_TAG or REMOVE_FROM_GROUP_TAG
     *
     * @param tags the tags based on which the messages will be filtered
     * @return List<BulletinBoardAPI.BulletinBoardMessage>
     */
    private List<BulletinBoardAPI.BulletinBoardMessage> GetRelevantMessages(List<String> tags) {
        BulletinBoardAPI.MessageFilterList.Builder filters =
                BulletinBoardAPI.MessageFilterList.newBuilder();

        for (String tag : tags) {
            BulletinBoardAPI.MessageFilter.Builder idFilter =
                    BulletinBoardAPI.MessageFilter.newBuilder().setTag(tag);

            filters.addFilter(idFilter);
        }
        return communicator.readMessages(filters.build());
    }

    /**
     * Gets Relevant bulletinBoard messages from communicator than converts them to VoterRegistryMessages
     *
     * @param tags list of tags that will be used as filters
     * @return List<VoterRegistryMessage>
     */
    private List<VoterRegistryMessage> GetRelevantVoterRegistryMessages(List<String> tags) throws InvalidProtocolBufferException {
        List<BulletinBoardAPI.BulletinBoardMessage> relevantMessages = GetRelevantMessages(tags);
        List<BulletinBoardAPI.UnsignedBulletinBoardMessage> messages =
                CollectionMessagesUtils.GetUnsignedBulletinBoardMessages(relevantMessages);
        return CollectionMessagesUtils.ConvertToVoterRegistryMessages(messages);
    }

    /**
     * Tries to post basicMessage and return true if its been successfully posted
     *
     * @param Message the massage to post
     * @return true when the post was successful else false
     */
    private boolean SafePost(BulletinBoardAPI.BulletinBoardMessage Message){
        try {
            communicator.postMessage(Message);
            return true;
        } catch (CommunicationException e) {
            return false;
        }
    }

    public void AddVoter(String voterID, String personalData, RegistryCallBack callback) {
        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder basicMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder().
                        addTag(RegistryTags.ID_TAG + " " + voterID)
                        .addTag(RegistryTags.VOTER_ENTRY_TAG.toString())
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        basicMessage.setData(ByteString.copyFrom(personalData.getBytes()));

        callback.HandleVoterAdded(SafePost(CreateBulletinBoardMessage(basicMessage.build())));
    }

    public void AddToGroup(String voterID, String groupID, RegistryCallBack callback) {
        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder basicMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + " " + voterID)
                        .addTag(RegistryTags.GROUP_ID_TAG + " " + groupID)
                        .addTag(RegistryTags.GROUP_ACTION_TAG + " " + RegistryTags.ADD_TO_GROUP_TAG)
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        callback.HandleVoterAddedToGroup(SafePost(CreateBulletinBoardMessage(basicMessage.build())));
    }

    public void RemoveFromGroup(String voterID, String groupID, RegistryCallBack callback) {
        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder basicMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + " " + voterID)
                        .addTag(RegistryTags.GROUP_ID_TAG + " " + groupID)
                        .addTag(RegistryTags.GROUP_ACTION_TAG + " " + RegistryTags.REMOVE_FROM_GROUP_TAG)
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        callback.HandleVoterRemovedFromGroup(SafePost(CreateBulletinBoardMessage(basicMessage.build())));
    }

    public void SetVoted(String id, RegistryCallBack callback) {
        BulletinBoardAPI.UnsignedBulletinBoardMessage.Builder basicMessage =
                BulletinBoardAPI.UnsignedBulletinBoardMessage.newBuilder()
                        .addTag(RegistryTags.ID_TAG + " " + id)
                        .addTag(RegistryTags.VOTE_ACTION_TAG.toString())
                        .addTag(RegistryTags.ACTION_TIMESTAMP_TAG + " " + AccurateTimestamp.GetCurrentTimestampString());

        callback.HandleVoterAddedToGroup(SafePost(CreateBulletinBoardMessage(basicMessage.build())));
    }

    public void GetGroups(String id, RegistryCallBack callback) {
        try {
            List<String> GroupsActionsTags = new ArrayList<String>(){{
                add(RegistryTags.ID_TAG + " " + id);
                add(RegistryTags.GROUP_ACTION_TAG + " " + RegistryTags.REMOVE_FROM_GROUP_TAG);
                add(RegistryTags.GROUP_ACTION_TAG + " " + RegistryTags.ADD_TO_GROUP_TAG);
            }};

            List<VoterRegistryMessage> voterRegistryMessages = null;
            voterRegistryMessages = GetRelevantVoterRegistryMessages(GroupsActionsTags);

            Map<String, VoterRegistryMessage> groupIdToMessage = CollectionMessagesUtils.GetLatestGroupsActions(voterRegistryMessages);
            callback.HandleVoterGroups(CollectionMessagesUtils.GetListOfGroupIds(groupIdToMessage));
        } catch (ParseException | InvalidProtocolBufferException e) {
            callback.HandleVoterGroups(null);
        }
    }

    public void GetPersonIDDetails(String id, RegistryCallBack callback) {
        try {
            List<String> GroupsActionsTags = new ArrayList<String>(){{
                add(RegistryTags.ID_TAG + " " + id);
                add(RegistryTags.VOTER_ENTRY_TAG.toString());
            }};

            List<VoterRegistryMessage> voterRegistryMessages =
                    GetRelevantVoterRegistryMessages(GroupsActionsTags);
            VoterRegistryMessage LatestMessage = voterRegistryMessages.get(0);


            // create FindLatestMessage
            for (VoterRegistryMessage message : voterRegistryMessages) {
                if (message.GetBasicMessageActionTimestamp().before(LatestMessage.GetBasicMessageActionTimestamp())) {
                    LatestMessage = message;
                }
            }
            // convert VoterRegistryMessage to list of strings 
            callback.HandleVoterInfo(Arrays.asList(
                    LatestMessage.GetWantedTagFromBasicMessage(RegistryTags.ID_TAG),
                    LatestMessage.base.getData().toString()));
        } catch (InvalidProtocolBufferException | ParseException e) {
            callback.HandleVoterInfo(null);
        }
    }
}
